--- pygrub.orig	2009-07-01 05:59:32.000000000 +0800
+++ pygrub.bsd	2009-07-01 06:00:05.000000000 +0800
@@ -90,6 +90,104 @@
 
     raise RuntimeError, "No root slice found"      
 
+# Some magic constants for the BSD disklabel
+BSD_LABEL_DISKMAGIC=0x82564557
+BSD_LABEL_LABELSECTOR=1
+BSD_LABEL_LABELOFFSET=0
+BSD_LABEL_MAXPARTITIONS=8
+BSD_LABEL_BBSIZE=8192
+BSD_LABEL_LABEL_PART=2
+BSD_LABEL_RAW_PART=2
+
+# Magic offsets in the label
+BSD_LABEL_OFFSET_MAGIC=0			# u_int32_t
+BSD_LABEL_OFFSET_SECTORSIZE=40			# u_int32_t - number of bytes per sector
+BSD_LABEL_OFFSET_PARTITIONTABLE=148		# structure * MAXPARTITIONS
+BSD_LABEL_OFFSET_NPARTITIONS=138		# u_int16_t
+
+# Magic offsets in the -partition- section
+BSD_LABEL_PARTITION_SIZE=0			# u_int32_t - size in sectors
+BSD_LABEL_PARTITION_OFFSET=4			# u_int32_t - offset in sectors
+BSD_LABEL_PARTITION_FSTYPE=12			# u_int8_t - partition type
+
+# How big is the partition structure?
+BSD_LABEL_PARTITION_STRUCT_SIZE=16
+
+# BSD label partition types that we care about
+BSD_FS_BSDFFS=7
+
+# Get the partition size from the given freebsd disklabel
+# partbuf contains the beginning of the disklabel
+def get_freebsd_partition_size(partbuf, partid):
+	i = BSD_LABEL_OFFSET_PARTITIONTABLE + (partid * BSD_LABEL_PARTITION_STRUCT_SIZE)
+	return struct.unpack("<I", partbuf[i+BSD_LABEL_PARTITION_SIZE:i+BSD_LABEL_PARTITION_SIZE+4])[0]
+
+def get_freebsd_partition_offset(partbuf, partid):
+	i = BSD_LABEL_OFFSET_PARTITIONTABLE + (partid * BSD_LABEL_PARTITION_STRUCT_SIZE)
+	return struct.unpack("<I", partbuf[i+BSD_LABEL_PARTITION_OFFSET:i+BSD_LABEL_PARTITION_OFFSET+4])[0]
+
+def get_freebsd_partition_fstype(partbuf, partid):
+	i = BSD_LABEL_OFFSET_PARTITIONTABLE + (partid * BSD_LABEL_PARTITION_STRUCT_SIZE)
+	return struct.unpack("<B", partbuf[i+BSD_LABEL_PARTITION_FSTYPE:i+BSD_LABEL_PARTITION_FSTYPE+1])[0]
+
+#
+# Find the root slice in a FreeBSD disklabel.
+#
+
+# The FreeBSD disklabel can either be dangerously dedicated or inside a DOS
+# partition. This function takes "offset" to mean the beginning of the FreeBSD
+# disk label and searches for a root partition "a" in the given slice.
+# The slice must have the FS type set to something that we know about - this
+# is currently "BSDFFS" (type 7). Anything else will fail.
+#
+# The filesystem must be at the beginning of that returned offset for the FS library
+# to discover and handle.
+
+# The returned offset is relative to the passed in offset - ie the beginning of
+# the "disklabel". This is so this code can be used regardless of the underlying
+# partitioning scheme (BSD FFS direct on disk, BSD FFS in a DOS primary partition,
+# etc.)
+
+def get_freebsd_slice(file, offset):
+    """Find the root slice in a FreeBSD disklabel."""
+    print "get_freebsd_slice: file:" , file
+    print "get_freebsd_slice: offset:" , offset
+
+    fd = os.open(file, os.O_RDONLY)
+    os.lseek(fd, offset + (SECTOR_SIZE * BSD_LABEL_LABELSECTOR) + BSD_LABEL_LABELOFFSET, 0)
+    buf = os.read(fd, 1024)
+
+    # Check the magic - is it a valid disklabel? if not, barf
+    bsd_d_magic = struct.unpack("<I", buf[BSD_LABEL_OFFSET_MAGIC:BSD_LABEL_OFFSET_MAGIC+4])[0]
+    print "bsd magic:", bsd_d_magic
+    print "needed magic: ", BSD_LABEL_DISKMAGIC
+    if bsd_d_magic <> BSD_LABEL_DISKMAGIC:
+    	raise RuntimeError, "FreeBSD label diskmagic incorrect"
+
+    # Get sectorsize - does it make sense? If not, barf
+    bsd_d_secsize = struct.unpack("<I", buf[BSD_LABEL_OFFSET_SECTORSIZE:BSD_LABEL_OFFSET_SECTORSIZE+4])[0];
+    print "bsd sector size:", bsd_d_secsize
+
+    # Get the first partition - is it the right type? If not, barf
+    bsd_d_npartitions = struct.unpack("<H", buf[BSD_LABEL_OFFSET_NPARTITIONS:BSD_LABEL_OFFSET_NPARTITIONS+2])[0];
+    print "bsd number of partitions:", bsd_d_npartitions
+
+    # Grab the filesystem offset in sectors
+    bsd_d_part_offset = get_freebsd_partition_offset(buf, 0)
+    bsd_d_part_size = get_freebsd_partition_size(buf, 0)
+    bsd_d_part_fstype = get_freebsd_partition_fstype(buf, 0)
+    print "bsd partition 0 offset:", bsd_d_part_offset
+    print "bsd partition 0 size:", bsd_d_part_size
+    print "bsd partition 0 type:", bsd_d_part_fstype
+
+    if bsd_d_part_fstype <> BSD_FS_BSDFFS:
+    	raise RuntimeError, "First BSD partition is not type BSDFFS (7)"
+
+    # XXX one should now verify that the given partition size and offset fall within the disk size..
+
+    # Calculate the filesystem offset, return that
+    return (bsd_d_part_offset * bsd_d_secsize)
+
 def get_fs_offset_gpt(file):
     fd = os.open(file, os.O_RDONLY)
     # assume the first partition is an EFI system partition.
@@ -100,6 +198,7 @@
 FDISK_PART_SOLARIS=0xbf
 FDISK_PART_SOLARIS_OLD=0x82
 FDISK_PART_GPT=0xee
+FDISK_PART_FREEBSD=0xa5
 
 def get_fs_offset(file):
     if not is_disk_image(file):
@@ -116,6 +215,9 @@
     if type == FDISK_PART_SOLARIS or type == FDISK_PART_SOLARIS_OLD:
         offset += get_solaris_slice(file, offset)
 
+    if type == FDISK_PART_FREEBSD:
+	offset += get_freebsd_slice(file, offset)
+
     if type == FDISK_PART_GPT:
         offset = get_fs_offset_gpt(file)
     
